# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import os
import sys
from io import StringIO

from requests import HTTPError, Session

from .client import CORREIOS_WEBSERVICES
from .utils import get_resource_path

WSDL_DIR = get_resource_path("wsdls")


class WSDLUpdater:
    webservices = CORREIOS_WEBSERVICES

    def __init__(self, wsdl_path: str, session=None, output=None) -> None:
        self.wsdl_path = wsdl_path

        if session is None:
            session = Session()
        self.session = session

        if output is None:
            output = StringIO()
        self.output = output

    def out(self, msg):
        print(msg, file=self.output)

    def save(self, path: str, content: str):
        with open(path, 'w') as file:
            file.write(content)
        self.out('Successfully create file: {}'.format(path))

    def download_wsdl(self, url: str, filename: str):
        path = os.path.join(self.wsdl_path, filename)

        self.out('Updating file: {} URL: {}'.format(filename, url))

        response = self.session.get(url)
        response.raise_for_status()

        self.save(path, response.text)
        self.out('Updated with success')

    def update_all(self):
        self.out('Updating Correios Webservice WSDLs')
        self.out('Files will be saved on: {}'.format(self.wsdl_path))

        for webservice, args in self.webservices.items():
            url, filename = args
            try:
                self.download_wsdl(url, filename)
            except HTTPError as ex:
                self.out("Error downloading {}: {}".format(url, ex))


def cli():
    parser = argparse.ArgumentParser(description='Updates Correios WSDL files')
    parser.add_argument(
        '-q',
        '--quiet',
        dest='output',
        action='store_const',
        const=None,
        default=sys.stdout,
    )
    parser.add_argument(
        '-p',
        '--path',
        dest='wsdl_path',
        default=str(WSDL_DIR),
        help='Custom path where wsdl files will be saved.'
    )
    args = parser.parse_args()

    updater = WSDLUpdater(wsdl_path=args.wsdl_path, output=args.output)
    return updater.update_all()


if __name__ == '__main__':
    sys.exit(cli() or 0)



# Copyright 2017 Adler Medrado
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ..correios.client import Correios
from ..correios.models.user import Service, User


def get_tracking_codes(service, quantity):
    olist_user = User("Your Company's Name", "Your Company's CNPJ")
    client = Correios("Your Correio's username", "Your correio's password")
    tracking_codes = client.request_tracking_codes(olist_user, Service.get(service), quantity=quantity)
    print(tracking_codes)


get_tracking_codes('41068', 1)  # Request 1 PAC Tracking Code
get_tracking_codes('40068', 1)  # Request 1 SEDEX Tracking Code



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from decimal import Decimal
from pathlib import Path
from typing import List, Optional, Sequence, Union

from requests.exceptions import Timeout
from zeep.exceptions import Fault

from .exceptions import (
    AuthenticationError,
    CanceledPostingCardError,
    ClientError,
    ClosePostingListError,
    ConnectTimeoutError,
    NonexistentPostingCardError,
    TrackingCodesLimitExceededError,
)
from .models.address import ZipAddress, ZipCode
from .models.builders import ModelBuilder
from .models.data import EXTRA_SERVICE_AR, EXTRA_SERVICE_MP
from .models.posting import FreightResponse, Package, PostingList, TrackingCode
from .models.user import ExtraService, PostingCard, Service, User
from .serializers import PostingListSerializer
from .soap import SoapClient
from .utils import get_resource_path

KG = 1000  # g
# environ servico url filename
CORREIOS_WEBSERVICES = {
    'sigep-production': (
        'https://apps.correios.com.br/SigepMasterJPA/AtendeClienteService/AtendeCliente?wsdl',
        'AtendeCliente-production.wsdl',
    ),
    'sigep-test': (
        'https://apphom.correios.com.br/SigepMasterJPA/AtendeClienteService/AtendeCliente?wsdl',
        'AtendeCliente-test.wsdl',
    ),
    'websro': (
        'https://webservice.correios.com.br/service/rastro/Rastro.wsdl',
        'Rastro.wsdl',
    ),
    'freight': (
        'http://ws.correios.com.br/calculador/CalcPrecoPrazo.asmx?WSDL',
        'CalcPrecoPrazo.asmx',
    ),
}

ERRORS = {
    re.compile(r"autenticacao"): AuthenticationError,
    re.compile(r"^O Cartão de Postagem.*Cancelado.$"): CanceledPostingCardError,
    re.compile(r"^Cartao de Postagem inexistente"): NonexistentPostingCardError,
}


class Correios:
    PRODUCTION = "production"
    TEST = "test"
    MAX_TRACKING_CODES_PER_REQUEST = 50

    def __init__(
        self,
        username: str,
        password: str,
        timeout: int = 8,
        environment: str = "production",
        local_wsdl_path: Optional[Path] = None
    ) -> None:

        if local_wsdl_path is None:
            local_wsdl_path = get_resource_path("wsdls")
        self.local_wsdl_path = local_wsdl_path

        ssl_check = environment == "production"
        sigep_key = "sigep-{}".format(environment)
        if not self.local_wsdl_path:
            sigep_url = (CORREIOS_WEBSERVICES[sigep_key][0], ssl_check)
            websro_url = CORREIOS_WEBSERVICES["websro"][0]
            freight_url = CORREIOS_WEBSERVICES["freight"][0]
        else:
            sigep_url = (str(self.local_wsdl_path / CORREIOS_WEBSERVICES[sigep_key][1]), ssl_check)
            websro_url = str(self.local_wsdl_path / CORREIOS_WEBSERVICES["websro"][1])
            freight_url = str(self.local_wsdl_path / CORREIOS_WEBSERVICES["freight"][1])

        self.username = username
        self.password = password
        self.timeout = timeout

        self.sigep_url = sigep_url[0]
        self.sigep_verify = sigep_url[1]

        self.sigep_client = SoapClient(self.sigep_url, verify=self.sigep_verify, timeout=self.timeout)
        self.sigep = self.sigep_client.service

        self.websro_client = SoapClient(websro_url, timeout=self.timeout)
        self.websro = self.websro_client.service

        self.freight_client = SoapClient(freight_url, timeout=self.timeout)
        self.freight = self.freight_client.service

        self.model_builder = ModelBuilder()

    def _handle_exception(self, exception):
        message = str(exception)

        for regex, exception in ERRORS.items():
            if regex.search(message):
                raise exception(message)

        raise ClientError(message)

    def _auth_call(self, method_name, *args, **kwargs):
        kwargs.update({
            "usuario": self.username,
            "senha": self.password,
        })
        return self._call(method_name, *args, **kwargs)

    def _call(self, method_name, *args, **kwargs):
        method = getattr(self.sigep, method_name)
        try:
            return method(*args, **kwargs)

        except Timeout:
            raise ConnectTimeoutError("Timeout connection error ({} seconds)".format(self.timeout))

        except Fault as exc:
            self._handle_exception(exc)

    def get_user(self, contract_number: Union[int, str], posting_card_number: Union[int, str]) -> User:
        contract_number = str(contract_number)
        posting_card_number = str(posting_card_number)
        user_data = self._auth_call("buscaCliente", contract_number, posting_card_number)
        return self.model_builder.build_user(user_data)

    def find_zipcode(self, zip_code: Union[ZipCode, str]) -> ZipAddress:
        zip_address_data = self._call("consultaCEP", str(zip_code))
        return self.model_builder.build_zip_address(zip_address_data)

    def verify_service_availability(
        self,
        posting_card: PostingCard,
        service: Service,
        from_zip_code: Union[ZipCode, str],
        to_zip_code: Union[ZipCode, str]
    ) -> bool:
        from_zip_code = ZipCode.create(from_zip_code)
        to_zip_code = ZipCode.create(to_zip_code)
        result = self._auth_call("verificaDisponibilidadeServico",
                                 posting_card.administrative_code, str(service),
                                 str(from_zip_code), str(to_zip_code))
        return result

    def get_posting_card_status(self, posting_card: PostingCard) -> bool:
        result = self._auth_call("getStatusCartaoPostagem", posting_card.number)
        return self.model_builder.build_posting_card_status(result)

    def request_tracking_codes(self, user: User, service: Service, quantity=1, receiver_type="C") -> list:
        result = self._auth_call("solicitaEtiquetas",
                                 receiver_type, str(user.federal_tax_number),
                                 service.id, quantity)
        return self.model_builder.build_tracking_codes_list(result)

    def generate_verification_digit(self, tracking_codes: Sequence[str]) -> List[int]:
        tracking_codes = [TrackingCode(tc).nodigit for tc in tracking_codes]
        result = self._auth_call("geraDigitoVerificadorEtiquetas",
                                 tracking_codes)

        return result

    def _generate_xml_string(self, posting_list: PostingList) -> str:
        posting_list_serializer = PostingListSerializer()
        xml = posting_list_serializer.serialize(posting_list)
        return xml.decode("ISO-8859-1")

    def close_posting_list(self, posting_list: PostingList, posting_card: PostingCard) -> PostingList:
        xml = self._generate_xml_string(posting_list)
        tracking_codes = posting_list.get_tracking_codes()

        try:
            id_ = self._auth_call("fechaPlpVariosServicos", xml,
                                  posting_list.custom_id, posting_card.number,
                                  tracking_codes)
        except ClientError as exc:
            if str(exc).startswith("A PLP não será fechada"):
                message = "Unable to close PLP. Tracking codes {} are already assigned to another PLP"
                message = message.format(tracking_codes)
                raise ClosePostingListError(message)
            raise

        posting_list.close_with_id(id_)
        return posting_list

    def get_tracking_code_events(self, tracking_list):
        if isinstance(tracking_list, (str, TrackingCode)):
            tracking_list = [tracking_list]

        if len(tracking_list) > Correios.MAX_TRACKING_CODES_PER_REQUEST:
            msg = '{} tracking codes requested exceeds the limit of {} stabilished by the Correios'
            msg = msg.format(len(tracking_list), Correios.MAX_TRACKING_CODES_PER_REQUEST)
            raise TrackingCodesLimitExceededError(msg)

        tracking_codes = {}
        for tracking_code in tracking_list:
            tracking_code = TrackingCode.create(tracking_code)
            tracking_codes[tracking_code.code] = tracking_code

        response = self.websro.buscaEventosLista(self.username, self.password, "L", "T", "101",
                                                 tuple(tracking_codes.keys()))
        return self.model_builder.load_tracking_events(tracking_codes, response)

    def calculate_freights(
        self,
        posting_card: PostingCard,
        services: List[Union[Service, int]],
        from_zip: Union[ZipCode, int, str], to_zip: Union[ZipCode, int, str],
        package: Package,
        value: Union[Decimal, float] = 0.00,
        extra_services: Optional[Sequence[Union[ExtraService, int]]] = None,
    ) -> List[FreightResponse]:

        administrative_code = posting_card.administrative_code
        services = [Service.get(s) for s in services]
        from_zip = ZipCode.create(from_zip)
        to_zip = ZipCode.create(to_zip)

        if extra_services is None:
            extra_services = []
        else:
            extra_services = [ExtraService.get(es) for es in extra_services]

        response = self.freight.CalcPrecoPrazo(
            administrative_code,
            self.password,
            ",".join(str(s) for s in services),
            str(from_zip),
            str(to_zip),
            package.weight / KG,
            package.package_type,
            package.length,
            package.height,
            package.width,
            package.diameter,
            "S" if EXTRA_SERVICE_MP in extra_services else "N",
            value,
            "S" if EXTRA_SERVICE_AR in extra_services else "N",
        )
        return self.model_builder.build_freights_list(response)

    def calculate_delivery_time(self, service: Union[Service, int], from_zip: Union[ZipCode, int, str],
                                to_zip: Union[ZipCode, int, str]) -> int:
        service = Service.get(service)
        from_zip = ZipCode.create(from_zip)
        to_zip = ZipCode.create(to_zip)

        response = self.freight.CalcPrazo(str(service), str(from_zip), str(to_zip))
        return response.cServico[0].PrazoEntrega



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from unittest import mock

import pytest

from correios.models.posting import PostingList
from correios.models.user import PostingCard

from .conftest import ShippingLabelFactory

try:
    from correios.renderers.pdf import PostingReportPDFRenderer
except ImportError:
    PostingReportPDFRenderer = None

TESTDIR = os.path.dirname(__file__)


@pytest.mark.skipif(not PostingReportPDFRenderer, reason="PDF generation support disabled")
def test_render_basic_shipping_label():
    shipping_labels_renderer = PostingReportPDFRenderer()
    shipping_labels = [ShippingLabelFactory.build() for _ in range(5)]

    for sl in shipping_labels:
        sl.weight_template = "Peso (g): <b>{!s}</b>"
        sl.invoice_template = "NF: {!s}"
        sl.order_template = "Ped.: <font size=6>{!s}</font>"
        sl.contract_number_template = "Contrato: <b>{!s}</b>"
        sl.service_name_template = "<b>{!s}</b>"
        sl.volume_template = "Volume: {!s}/{!s}"
        shipping_labels_renderer.add_shipping_label(sl)

    pdf = shipping_labels_renderer.render_labels()
    assert bytes(pdf).startswith(b"%PDF-1.4")


@pytest.mark.skipif(not PostingReportPDFRenderer, reason="PDF generation support disabled")
def test_render_basic_posting_list(posting_list: PostingList, posting_card: PostingCard):
    posting_list.close_with_id(number=12345)
    shipping_labels_renderer = PostingReportPDFRenderer()
    shipping_labels = [ShippingLabelFactory.build(posting_card=posting_card) for _ in range(5)]
    for shipping_label in shipping_labels:
        posting_list.add_shipping_label(shipping_label)
    shipping_labels_renderer.set_posting_list(posting_list)
    pdf = shipping_labels_renderer.render_posting_list()
    assert bytes(pdf).startswith(b"%PDF-1.4")


@pytest.mark.skipif(not PostingReportPDFRenderer, reason="PDF generation support disabled")
def test_render_all_posting_docs(posting_list: PostingList, posting_card: PostingCard):
    posting_list.close_with_id(number=12345)
    shipping_labels_renderer = PostingReportPDFRenderer()
    shipping_labels = [ShippingLabelFactory.build(posting_card=posting_card) for _ in range(35)]
    for shipping_label in shipping_labels:
        posting_list.add_shipping_label(shipping_label)
    shipping_labels_renderer.set_posting_list(posting_list)
    pdf = shipping_labels_renderer.render()
    assert bytes(pdf).startswith(b"%PDF-1.4")


@pytest.mark.skipif(not PostingReportPDFRenderer, reason="PDF generation support disabled")
@mock.patch('correios.renderers.pdf.Canvas.drawString')
@pytest.mark.parametrize('display, count', (
    (True, 2),
    (False, 1),
))
def test_display_on_label_render_condition(mock_drawstring, display, count):
    shipping_labels_renderer = PostingReportPDFRenderer()
    shipping_labels = ShippingLabelFactory.build()

    shipping_labels.extra_services[0].display_on_label = display
    shipping_labels_renderer.add_shipping_label(shipping_labels)
    shipping_labels_renderer.render_labels()

    assert mock_drawstring.call_count == count



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from datetime import datetime, timedelta
from decimal import Decimal

import pytest
from PIL.Image import Image

from correios import exceptions
from correios.models import posting
from correios.models.data import (
    EXTRA_SERVICE_AR,
    EXTRA_SERVICE_RR,
    EXTRA_SERVICE_VD,
    SERVICE_PAC,
    SERVICE_SEDEX,
    TRACKING_EVENT_TYPES,
)
from correios.models.user import ExtraService, Service

from .conftest import ShippingLabelFactory

FIXTURESDIR = os.path.join(os.path.dirname(__file__), "fixtures")


@pytest.mark.parametrize("tracking_code", [
    "DL74668653 BR",
    "DL746686536BR",
    "DL74668653BR",
    "dl74668653br",
])
def test_tracking_code_constructor(tracking_code):
    tracking = posting.TrackingCode(tracking_code)
    assert str(tracking) == "DL746686536BR"
    assert tracking.code == "DL746686536BR"
    assert tracking.prefix == "DL"
    assert tracking.number == "74668653"
    assert tracking.digit == 6
    assert tracking.nodigit == "DL74668653 BR"
    assert tracking.short == "DL74668653BR"
    assert tracking.splitted == "DL 746 686 536 BR"
    assert repr(tracking) == "<TrackingCode code='DL746686536BR'>"


@pytest.mark.parametrize("tracking_code", [
    "DL7466865BR",
    "DL746686530BR",  # invalid digit (0)
    "DL7466X653 BR",
    "DL74668653B",
    "D746686530 BR",
    "DL46686530 B1",
])
def test_fail_invalid_tracking_code(tracking_code):
    with pytest.raises(exceptions.InvalidTrackingCodeError):
        posting.TrackingCode(tracking_code)


@pytest.mark.parametrize("tracking_code,digit", [
    ("DL74668653 BR", 6),
    ("DL02000000 BR", 0),
    ("DL00000000 BR", 5),
])
def test_tracking_code_digit_calculator(tracking_code, digit):
    tracking = posting.TrackingCode(tracking_code)
    assert tracking.digit == digit


def test_tracking_code_creator():
    tracking_code1 = posting.TrackingCode.create("DL746686536BR")
    assert tracking_code1.code == "DL746686536BR"

    tracking_code2 = posting.TrackingCode.create(tracking_code1)
    assert tracking_code1 == tracking_code2


def test_tracking_code_range_generator():
    tracking_codes = posting.TrackingCode.create_range("DL74668650 BR", "DL74668654 BR")
    assert len(tracking_codes) == 5
    assert all(isinstance(tc, posting.TrackingCode) for tc in tracking_codes)


def test_fail_tracking_code_invalid_range_generator():
    with pytest.raises(exceptions.InvalidTrackingCodeError):
        posting.TrackingCode.create_range("DL74668650 BR", "SX74668654 BR")  # different prefix

    with pytest.raises(exceptions.InvalidTrackingCodeError):
        posting.TrackingCode.create_range("DL74668650 BR", "DL74668654 US")  # different suffix

    with pytest.raises(exceptions.InvalidTrackingCodeError):
        posting.TrackingCode.create_range("DL74668654 BR", "DL74668650 BR")  # end < start


def test_basic_shipping_label(posting_card, sender_address, receiver_address, tracking_code, package):
    shipping_label = posting.ShippingLabel(
        posting_card=posting_card,
        sender=sender_address,
        receiver=receiver_address,
        service=SERVICE_SEDEX,
        extra_services=[EXTRA_SERVICE_AR],
        tracking_code=tracking_code,
        logo=os.path.join(FIXTURESDIR, "test_logo.jpg"),
        order="123",
        invoice_number="321",
        invoice_series="A1",
        invoice_type="",
        package=package,
        text="Hello World!",
        latitude=-25.4131980,
        longitude=-49.2584896,
    )

    assert shipping_label.posting_card == posting_card
    assert shipping_label.sender == sender_address
    assert shipping_label.receiver == receiver_address

    assert shipping_label.service == Service.get(SERVICE_SEDEX)
    assert shipping_label.get_service_name() == "SEDEX"

    assert shipping_label.extra_services == [ExtraService.get(EXTRA_SERVICE_RR),
                                             ExtraService.get(EXTRA_SERVICE_AR)]

    assert shipping_label.tracking_code == tracking_code
    assert shipping_label.get_tracking_code().replace(" ", "") == str(shipping_label.tracking_code)

    assert isinstance(shipping_label.logo, Image)

    # noinspection PyUnresolvedReferences
    assert shipping_label.logo.filename == os.path.join(FIXTURESDIR, "test_logo.jpg")

    assert shipping_label.order == "123"
    assert shipping_label.get_order() == "123"

    assert shipping_label.invoice_number == "321"
    assert shipping_label.get_invoice() == "321"

    assert shipping_label.get_contract_number() == "9911222777"
    assert shipping_label.get_package_sequence() == "{}/{}".format(*shipping_label.package.sequence)
    assert shipping_label.get_weight() == "{}g".format(shipping_label.package.weight)
    assert shipping_label.package.posting_list_volumetric_weight == Decimal("0.00")

    assert shipping_label.text == "Hello World!"

    assert shipping_label.latitude == -25.4131980
    assert shipping_label.longitude == -49.2584896

    assert shipping_label.posting_list_group == 0

    assert shipping_label.get_symbol_filename().endswith("/express.gif")
    assert isinstance(shipping_label.symbol, Image)

    assert len(shipping_label.get_datamatrix_info()) == 164  # datamatrix info size accordingly with documentation
    assert shipping_label.get_sender_data().count("<br/>") >= 2
    assert shipping_label.get_receiver_data().count("<br/>") >= 2

    assert repr(shipping_label) == "<ShippingLabel tracking='{!s}'>".format(shipping_label.tracking_code)


def test_basic_default_shipping_label(posting_card, sender_address, receiver_address, package):
    shipping_label = posting.ShippingLabel(
        posting_card=posting_card,
        sender=sender_address,
        receiver=receiver_address,
        service=4162,  # SERVICE_SEDEX_CODE
        package=package,
        tracking_code="PD12345678 BR",
    )

    assert shipping_label.service == Service.get(SERVICE_SEDEX)
    assert shipping_label.tracking_code.code == "PD123456785BR"
    assert len(shipping_label.extra_services) == 1


def test_shipping_label_with_declared_value(posting_card, sender_address, receiver_address, package):
    service = Service.get(SERVICE_SEDEX)
    shipping_label = posting.ShippingLabel(
        posting_card=posting_card,
        sender=sender_address,
        receiver=receiver_address,
        service=service,
        package=package,
        tracking_code="PD12345678 BR",
        value=service.max_declared_value - Decimal("1.00"),
        extra_services=[EXTRA_SERVICE_VD],
    )
    assert ExtraService.get(EXTRA_SERVICE_VD) in shipping_label.extra_services


def test_shipping_label_with_min_declared_value_pac(posting_card, sender_address, receiver_address, package):
    service = Service.get(SERVICE_PAC)
    shipping_label = posting.ShippingLabel(
        posting_card=posting_card,
        sender=sender_address,
        receiver=receiver_address,
        service=service,
        package=package,
        tracking_code="PD12345678 BR",
        value=Decimal("0"),
        extra_services=[EXTRA_SERVICE_VD],
    )
    assert shipping_label.value == Decimal("18.50")


def test_shipping_label_with_min_declared_value_sedex(posting_card, sender_address, receiver_address, package):
    service = Service.get(SERVICE_SEDEX)
    shipping_label = posting.ShippingLabel(
        posting_card=posting_card,
        sender=sender_address,
        receiver=receiver_address,
        service=service,
        package=package,
        tracking_code="PD12345678 BR",
        value=Decimal("0"),
        extra_services=[EXTRA_SERVICE_VD],
    )
    assert shipping_label.value == Decimal("18.50")


def test_fail_shipping_label_with_invalid_declared_value(posting_card, sender_address, receiver_address, package):
    service = Service.get(SERVICE_SEDEX)
    shipping_label = posting.ShippingLabel(
        posting_card=posting_card,
        sender=sender_address,
        receiver=receiver_address,
        service=service,
        package=package,
        tracking_code="PD12345678 BR",
        value=service.max_declared_value + Decimal("1.00"),
    )

    with pytest.raises(exceptions.MaximumDeclaredValueError):
        shipping_label.add_extra_service(ExtraService.get(EXTRA_SERVICE_VD))


def test_fail_shipping_label_same_addresses(posting_card, sender_address, tracking_code, package):
    with pytest.raises(exceptions.InvalidAddressesError):
        posting.ShippingLabel(
            posting_card,
            sender_address,
            sender_address,
            SERVICE_SEDEX,
            package=package,
            tracking_code=tracking_code,
        )


def test_package_basic():
    package = posting.Package(
        package_type=posting.Package.TYPE_BOX,
        width=11,
        height=10,
        length=16,
        weight=10000,
        service=SERVICE_PAC,
    )
    assert isinstance(package.service, Service)
    assert package.package_type == posting.Package.TYPE_BOX


@pytest.mark.parametrize("package,freight_package_type", [
    (posting.Package(posting.Package.TYPE_ENVELOPE, 0, 0, 0, weight=1), 3),
    (posting.Package(posting.Package.TYPE_BOX, 11, 10, 16, weight=1), 1),
    (posting.Package(posting.Package.TYPE_CYLINDER, 0, 0, 14, 2, weight=1), 2),
])
def test_freight_package_type(package, freight_package_type):
    assert package.freight_package_type == freight_package_type


def test_package_basic_envelop_dimensions_validation():
    posting.Package.validate(
        posting.Package.TYPE_ENVELOPE,
        0, 0, 0, 0,
        weight=1,
    )


@pytest.mark.parametrize("weight,width,height,length,posting_weight", [
    (9000, 50, 60, 15, 9000),
    (15000, 43, 28, 52, 15000),
    (7000, 55, 31, 40, 11367),
    (15000, 73, 73, 73, 64837)  # math.ceil(64836.1)
])
def test_package_posting_weight_calculation(weight, width, height, length, posting_weight):
    volumetric_weight = posting.Package.calculate_volumetric_weight(width, height, length)
    assert posting.Package.calculate_posting_weight(weight, volumetric_weight) == posting_weight


@pytest.mark.parametrize("package_type,width,height,length,diameter,exc", [
    (posting.Package.TYPE_ENVELOPE, 1, 0, 0, 0, exceptions.InvalidPackageDimensionsError),
    (posting.Package.TYPE_ENVELOPE, 0, 1, 0, 0, exceptions.InvalidPackageDimensionsError),
    (posting.Package.TYPE_ENVELOPE, 0, 0, 1, 0, exceptions.InvalidPackageDimensionsError),
    (posting.Package.TYPE_ENVELOPE, 0, 0, 0, 1, exceptions.InvalidPackageDimensionsError),
    (posting.Package.TYPE_ENVELOPE, 1, 1, 1, 1, exceptions.InvalidPackageDimensionsError),
    (posting.Package.TYPE_BOX, 11, 2, 16, 1, exceptions.InvalidPackageDimensionsError),  # invalid diameter
    (posting.Package.TYPE_BOX, 110, 2, 16, 0, exceptions.InvalidMaxPackageDimensionsError),  # max width=105
    (posting.Package.TYPE_BOX, 11, 110, 16, 0, exceptions.InvalidMaxPackageDimensionsError),  # max height=110
    (posting.Package.TYPE_BOX, 11, 2, 110, 0, exceptions.InvalidMaxPackageDimensionsError),  # max length=110
    (posting.Package.TYPE_BOX, 105, 105, 105, 0, exceptions.InvalidMaxPackageDimensionsError),  # sum > 200
    (posting.Package.TYPE_CYLINDER, 1, 0, 18, 16, exceptions.InvalidPackageDimensionsError),  # invalid width
    (posting.Package.TYPE_CYLINDER, 0, 1, 18, 16, exceptions.InvalidPackageDimensionsError),  # invalid height
    (posting.Package.TYPE_CYLINDER, 0, 0, 110, 16, exceptions.InvalidMaxPackageDimensionsError),  # max length=105
    (posting.Package.TYPE_CYLINDER, 0, 0, 18, 110, exceptions.InvalidMaxPackageDimensionsError),  # max diameter=91
    (posting.Package.TYPE_CYLINDER, 0, 0, 30, 90, exceptions.InvalidMaxPackageDimensionsError),  # max cylinder size=200
])
def test_fail_package_dimensions_validation(package_type, width, height, length, diameter, exc):
    with pytest.raises(exc):
        posting.Package.validate(package_type, width, height, length, diameter, weight=1)


def test_package_weight_validation():
    posting.Package.validate(
        posting.Package.TYPE_BOX,
        12, 10, 20,
        service=Service.get(SERVICE_SEDEX),
        weight=10000,
    )

    posting.Package.validate(
        posting.Package.TYPE_BOX,
        12, 10, 20,
        service=Service.get(10065),  # 10065 - service with no max weight
        weight=50000,
    )


def test_fail_package_weight_validation():
    with pytest.raises(exceptions.InvalidMaxPackageWeightError):
        posting.Package.validate(
            posting.Package.TYPE_BOX,
            12, 10, 20,
            service=Service.get(SERVICE_SEDEX),
            weight=50000,
        )


def test_fix_bug_of_weight_using_diameter_information():
    package = posting.Package(
        package_type=posting.Package.TYPE_CYLINDER,
        diameter=2,
        length=20,
        weight=10000,
        service=SERVICE_PAC,
    )
    assert package.real_weight == 10000
    assert package.real_diameter == 2


@pytest.mark.parametrize('package_type,diameter,result', [
    (posting.Package.TYPE_ENVELOPE, 16, 0),
    (posting.Package.TYPE_BOX, 18, 0),
    (posting.Package.TYPE_CYLINDER, 3, 5),
    (posting.Package.TYPE_CYLINDER, 18, 18),
])
def test_package_diameter(package, package_type, diameter, result):
    package.package_type = package_type
    package.real_diameter = diameter
    assert package.diameter == result


@pytest.mark.parametrize("sequence", [
    (1,),
    (3, 2),
])
def test_fail_package_invalid_sequence(sequence):
    with pytest.raises(exceptions.InvalidPackageSequenceError):
        posting.Package(
            package_type=posting.Package.TYPE_BOX,
            width=11,
            height=10,
            length=16,
            weight=10000,
            sequence=sequence,  # invalid tuple
        )


@pytest.mark.parametrize("dimension,value,minimum", [
    ("width", 1, posting.MIN_WIDTH),
    ("height", 1, posting.MIN_HEIGHT),
    ("length", 1, posting.MIN_LENGTH),
    ("weight", 1, 1),
])
def test_box_package_change_dimensions_below_minimum(package, dimension, value, minimum):
    setattr(package, dimension, value)
    assert getattr(package, dimension) == minimum


@pytest.mark.parametrize("dimension,value,minimum", [
    ("diameter", 1, posting.MIN_DIAMETER),
    ("length", 1, posting.MIN_LENGTH),
    ("weight", 1, 1),
])
def test_cylinder_package_change_dimensions_below_minimum(package, dimension, value, minimum):
    package.package_type = package.TYPE_CYLINDER
    setattr(package, dimension, value)
    assert getattr(package, dimension) == minimum


@pytest.mark.parametrize("dimension,value,exc", [
    ("width", 0, exceptions.InvalidMinPackageDimensionsError),
    ("width", posting.MAX_WIDTH + 1, exceptions.InvalidMaxPackageDimensionsError),
    ("height", 0, exceptions.InvalidMinPackageDimensionsError),
    ("height", posting.MAX_HEIGHT + 1, exceptions.InvalidMaxPackageDimensionsError),
    ("length", 0, exceptions.InvalidMinPackageDimensionsError),
    ("length", posting.MAX_LENGTH + 1, exceptions.InvalidMaxPackageDimensionsError),
    ("weight", 0, exceptions.InvalidMinPackageWeightError),
    ("weight", 100000, exceptions.InvalidMaxPackageWeightError),
])
def test_fail_box_package_change_invalid_dimensions(package, dimension, value, exc):
    with pytest.raises(exc):
        setattr(package, dimension, value)


@pytest.mark.parametrize("dimension,value,exc", [
    ("diameter", 0, exceptions.InvalidMinPackageDimensionsError),
    ("diameter", posting.MAX_DIAMETER + 1, exceptions.InvalidMaxPackageDimensionsError),
    ("length", 0, exceptions.InvalidMinPackageDimensionsError),
    ("length", posting.MAX_LENGTH + 1, exceptions.InvalidMaxPackageDimensionsError),
])
def test_fail_cylinder_package_change_invalid_dimensions(package, dimension, value, exc):
    package.package_type = package.TYPE_CYLINDER
    with pytest.raises(exc):
        setattr(package, dimension, value)


def test_basic_posting_list(shipping_label):
    posting_list = posting.PostingList(custom_id=12345)
    posting_list.add_shipping_label(shipping_label)

    assert posting_list.custom_id == 12345
    assert not posting_list.closed
    tracking_codes = posting_list.get_tracking_codes()
    assert tracking_codes and shipping_label.tracking_code.short in tracking_codes


def test_fail_add_different_sender_in_posting_list():
    posting_list = posting.PostingList(custom_id=12345)
    posting_list.add_shipping_label(ShippingLabelFactory.build())

    with pytest.raises(exceptions.PostingListError):
        posting_list.add_shipping_label(ShippingLabelFactory.build())


def test_fail_add_same_shipping_label_twice_in_posting_list(shipping_label):
    posting_list = posting.PostingList(custom_id=12345)
    posting_list.add_shipping_label(shipping_label)

    with pytest.raises(exceptions.PostingListError):
        posting_list.add_shipping_label(shipping_label)


def test_calculate_insurance_when_not_applicable():
    value = posting.Package.calculate_insurance(per_unit_value=50, quantity=2, service=SERVICE_SEDEX)
    assert value == Decimal(0)

    value = posting.Package.calculate_insurance(per_unit_value=Decimal(10), service=SERVICE_PAC)
    assert value == Decimal(0)


def test_calculate_insurance_pac():
    value = posting.Package.calculate_insurance(per_unit_value=193, service=SERVICE_PAC)
    assert value == Decimal(1)

    value = posting.Package.calculate_insurance(per_unit_value=Decimal(193), quantity=2, service=SERVICE_PAC)
    assert value == Decimal(2)

    value = posting.Package.calculate_insurance(per_unit_value=Decimal(500), quantity=2, service=SERVICE_PAC)
    assert value == Decimal('6.30')


def test_calculate_insurance_sedex():
    value = posting.Package.calculate_insurance(per_unit_value=Decimal(500), service=SERVICE_SEDEX)
    assert value == Decimal('2.98')

    value = posting.Package.calculate_insurance(per_unit_value=Decimal(500), quantity=2, service=SERVICE_SEDEX)
    assert value == Decimal('5.95')


def test_event_status():
    event_status = posting.EventStatus('BDE', 1)
    assert repr(event_status) == "<EventStatus('BDE', 1)>"


@pytest.mark.parametrize("status", (posting.EventStatus("BDE", "01"), ("BDE", "01")))
def test_basic_tracking_event(status):
    tracking_event = posting.TrackingEvent(
        timestamp=datetime(2010, 1, 2, 1, 2),
        status=status,
        location_zip_code="82940150",
        location="Correios",
        receiver="José",
        city="Curitiba",
        state="PR",
        document="XYZ",
        comment="The comment",
        description="The description",
        details="The details",
    )

    assert tracking_event.timestamp == datetime(2010, 1, 2, 1, 2)
    assert tracking_event.status.type == "BDE"
    assert tracking_event.status.status == 1
    assert tracking_event.location_zip_code == "82940150"
    assert tracking_event.location == "Correios"
    assert tracking_event.receiver == "José"
    assert tracking_event.city == "Curitiba"
    assert tracking_event.state == "PR"
    assert tracking_event.document == "XYZ"
    assert tracking_event.comment == "The comment"
    assert tracking_event.description == "The description"
    assert tracking_event.details == "The details"

    assert repr(tracking_event) == "<TrackingEvent((BDE, 1), 02/01/2010 01:02)>"
    assert str(tracking_event) == "The description - Correios - Curitiba/PR"


def test_tracking_event_timestamp_format(tracking_event):
    expected_date = "01/01/2016 12:00"
    assert tracking_event.timestamp.strftime(posting.TrackingEvent.timestamp_format) == expected_date


def test_basic_not_found_tracking_event():
    tracking_event = posting.NotFoundTrackingEvent(
        timestamp=datetime(2010, 1, 2, 1, 2),
        comment="Not found",
    )
    assert tracking_event.timestamp == datetime(2010, 1, 2, 1, 2)
    assert tracking_event.status.type == "ERROR"
    assert tracking_event.status.status == 0
    assert tracking_event.comment == "Not found"


@pytest.mark.parametrize("status_type,status_number", [
    ("BDE", 0),
    ("BDI", 0),
    ("BDR", 0),
    ("BLQ", 1),
    # ("CAR", 0),
    ("CD", 0),
    ("CMT", 0),
    ("CO", 1),
    ("CUN", 0),
    ("DO", 0),
    ("EST", 1),
    ("FC", 1),
    ("IDC", 1),
    ("LDI", 0),
    ("LDE", 0),
    ("OEC", 0),
    ("PAR", 15),
    ("PMT", 1),
    ("PO", 0),
    ("RO", 0),
    ("TRI", 0),
    # ("CMR", 0),
])
def test_basic_event_status(status_type, status_number):
    event_status = posting.EventStatus(status_type, status_number)

    assert event_status.type == status_type
    assert event_status.status == status_number
    assert event_status.display_event_type == TRACKING_EVENT_TYPES[status_type]

    assert str(event_status) == "({}, {})".format(status_type, status_number)
    assert repr(event_status) == "<EventStatus({!r}, {!r})>".format(status_type, status_number)


@pytest.mark.parametrize("event_type", ("XYZ", "ABC", "XXX", "WTF", "BD", "ERROR"))
def test_invalid_event_status(event_type):
    with pytest.raises(exceptions.InvalidEventStatusError):
        posting.EventStatus(event_type, 1)


def test_basic_freight():
    freight = posting.FreightResponse(SERVICE_SEDEX, timedelta(days=5), Decimal("10.00"))
    assert freight.total == Decimal("10.00")
    assert freight.delivery_time == timedelta(days=5)
    assert freight.declared_value == Decimal("0.00")
    assert freight.saturday is False
    assert freight.home is False


def test_basic_freight_conversion():
    freight = posting.FreightResponse(SERVICE_SEDEX, 5, 10.00)
    assert freight.delivery_time == timedelta(days=5)
    assert freight.total == Decimal("10.00")


@pytest.mark.parametrize('package_type,width,height,length,diameter,result', [
    (posting.Package.TYPE_BOX, 11, 10, 16, 0, True),
    (posting.Package.TYPE_BOX, 70, 10, 10, 0, True),
    (posting.Package.TYPE_BOX, 10, 70, 10, 0, True),
    (posting.Package.TYPE_BOX, 10, 10, 70, 0, True),
    (posting.Package.TYPE_BOX, 71, 10, 10, 0, False),
    (posting.Package.TYPE_BOX, 10, 71, 10, 0, False),
    (posting.Package.TYPE_BOX, 10, 10, 71, 0, False),
    (posting.Package.TYPE_CYLINDER, 0, 0, 14, 2, False),
])
def test_package_is_mechanizable(package_type, width, height, length, diameter, result):
    package = posting.Package(package_type, width, height, length, diameter, weight=1)
    assert package.is_mechanizable == result


@pytest.mark.parametrize('package_type,width,height,length,diameter,cost', [
    (posting.Package.TYPE_BOX, 11, 10, 16, 0, 0),
    (posting.Package.TYPE_BOX, 70, 10, 10, 0, 0),
    (posting.Package.TYPE_BOX, 10, 70, 10, 0, 0),
    (posting.Package.TYPE_BOX, 10, 10, 70, 0, 0),
    (posting.Package.TYPE_BOX, 71, 10, 10, 0, posting.NON_MECHANIZABLE_COST),
    (posting.Package.TYPE_BOX, 10, 71, 10, 0, posting.NON_MECHANIZABLE_COST),
    (posting.Package.TYPE_BOX, 10, 10, 71, 0, posting.NON_MECHANIZABLE_COST),
    (posting.Package.TYPE_CYLINDER, 0, 0, 14, 2, posting.NON_MECHANIZABLE_COST),
])
def test_package_non_mechanizable_cost(package_type, width, height, length, diameter, cost):
    package = posting.Package(package_type, width, height, length, diameter, weight=1)
    assert package.non_mechanizable_cost == cost



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
from typing import Dict

from ..utils import to_decimal, to_integer
from .address import ZipAddress
from .posting import EventStatus, FreightResponse, NotFoundTrackingEvent, TrackingCode, TrackingEvent
from .user import Contract, FederalTaxNumber, PostingCard, Service, StateTaxNumber, User


class ModelBuilder:
    def build_service(self, service_data):
        service = Service(
            code=service_data.codigo,
            id=service_data.id,
            description=service_data.descricao,
            category=service_data.servicoSigep.categoriaServico
        )
        return service

    def build_posting_card(self, contract: Contract, posting_card_data):
        posting_card = PostingCard(
            contract=contract,
            number=posting_card_data.numero,
            administrative_code=posting_card_data.codigoAdministrativo,
        )

        posting_card.start_date = posting_card_data.dataVigenciaInicio
        posting_card.end_date = posting_card_data.dataVigenciaFim
        posting_card.status = posting_card_data.statusCartaoPostagem
        posting_card.status_code = posting_card_data.statusCodigo
        posting_card.unit = posting_card_data.unidadeGenerica

        for service_data in posting_card_data.servicos:
            service = self.build_service(service_data)
            posting_card.add_service(service)

        return posting_card

    def build_contract(self, user: User, contract_data):
        contract = Contract(
            user=user,
            number=contract_data.contratoPK.numero,
            regional_direction=contract_data.codigoDiretoria,
        )

        contract.customer_code = contract_data.codigoCliente
        contract.status_code = contract_data.statusCodigo
        contract.start_date = contract_data.dataVigenciaInicio
        contract.end_date = contract_data.dataVigenciaFim

        for posting_card_data in contract_data.cartoesPostagem:
            self.build_posting_card(contract, posting_card_data)

        return contract

    def build_user(self, user_data):
        user = User(
            name=user_data.nome,
            federal_tax_number=FederalTaxNumber(user_data.cnpj),
            state_tax_number=StateTaxNumber(user_data.inscricaoEstadual),
            status_number=user_data.statusCodigo,
        )

        for contract_data in user_data.contratos:
            self.build_contract(user, contract_data)

        return user

    def build_zip_address(self, zip_address_data):
        zip_address = ZipAddress(
            id=zip_address_data.id,
            zip_code=zip_address_data.cep,
            state=zip_address_data.uf,
            city=zip_address_data.cidade,
            district=zip_address_data.bairro,
            address=zip_address_data.end,
            complements=[zip_address_data.complemento, zip_address_data.complemento2]
        )
        return zip_address

    def build_posting_card_status(self, response):
        if response.lower() != "normal":
            return PostingCard.CANCELLED
        return PostingCard.ACTIVE

    def build_tracking_codes_list(self, response):
        codes = response.split(",")
        return TrackingCode.create_range(codes[0], codes[1])

    def _load_invalid_event(self, tracking_code: TrackingCode, tracked_object):
        event = NotFoundTrackingEvent(
            timestamp=datetime.now(),
            comment=tracked_object.erro,
        )
        tracking_code.add_event(event)

    def _load_events(self, tracking_code: TrackingCode, events):
        for event in events:
            timestamp = datetime.strptime("{} {}".format(event.data, event.hora), TrackingEvent.timestamp_format)
            event = TrackingEvent(
                timestamp=timestamp,
                status=EventStatus(event.tipo, event.status),
                location_zip_code=getattr(event, "codigo", ""),
                location=getattr(event, "local", ""),
                city=getattr(event, "cidade", ""),
                state=getattr(event, "uf", ""),
                receiver=getattr(event, "recebedor", ""),
                document=getattr(event, "documento", ""),
                comment=getattr(event, "comentario", ""),
                description=getattr(event, "descricao", ""),
                details=getattr(event, "detalhes", ""),
            )

            tracking_code.add_event(event)

    def load_tracking_events(self, tracking_codes: Dict[str, TrackingCode], response):
        result = []
        for tracked_object in response.objeto:
            tracking_code = tracking_codes[tracked_object.numero]

            if 'erro' in tracked_object:
                self._load_invalid_event(tracking_code, tracked_object)
            else:
                tracking_code.name = tracked_object.nome
                tracking_code.initials = tracked_object.sigla
                tracking_code.category = tracked_object.categoria
                self._load_events(tracking_code, tracked_object.evento)

            result.append(tracking_code)

        return result

    def build_freights_list(self, response):
        result = []
        for service_data in response.cServico:
            freight = self.build_freight(service_data=service_data)
            result.append(freight)
        return result

    def build_freight(self, service_data) -> FreightResponse:
        return FreightResponse(
            service=Service.get(service_data.Codigo),
            delivery_time=int(service_data.PrazoEntrega),
            value=to_decimal(service_data.ValorSemAdicionais),
            declared_value=to_decimal(service_data.ValorValorDeclarado),
            mp_value=to_decimal(service_data.ValorMaoPropria),
            ar_value=to_decimal(service_data.ValorAvisoRecebimento),
            saturday=service_data.EntregaSabado and service_data.EntregaSabado.lower() == "s" or False,
            home=service_data.EntregaDomiciliar and service_data.EntregaDomiciliar.lower() == "s" or False,
            error_code=to_integer(service_data.Erro) or 0,
            error_message=service_data.MsgErro or ""
        )



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from decimal import Decimal
from unittest import mock

import pytest
from requests.exceptions import ConnectTimeout
from zeep.exceptions import Fault

from correios.exceptions import (
    AuthenticationError,
    CanceledPostingCardError,
    ClosePostingListError,
    ConnectTimeoutError,
    NonexistentPostingCardError,
    PostingListSerializerError,
    TrackingCodesLimitExceededError,
)
from correios.models.address import ZipCode
from correios.models.builders import ModelBuilder
from correios.models.data import (
    EXTRA_SERVICE_AR,
    EXTRA_SERVICE_MP,
    EXTRA_SERVICE_VD,
    FREIGHT_ERROR_FINAL_ZIPCODE_RESTRICTED,
    FREIGHT_ERROR_INITIAL_AND_FINAL_ZIPCODE_RESTRICTED,
    FREIGHT_ERROR_INITIAL_ZIPCODE_RESTRICTED,
    SERVICE_PAC,
    SERVICE_SEDEX,
    SERVICE_SEDEX10,
)
from correios.models.posting import FreightResponse, NotFoundTrackingEvent, TrackingCode
from correios.models.user import ExtraService, PostingCard, Service
from correios.serializers import PostingListSerializer
from correios.utils import get_resource_path

from .vcr import vcr

try:
    from correios import client as correios
except ImportError:
    correios = None


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_basic_client():
    client = correios.Correios(username="sigep", password="XXXXXX", environment=correios.Correios.TEST)
    assert client.sigep_url == str(get_resource_path('wsdls/AtendeCliente-test.wsdl'))
    assert not client.sigep_verify
    assert client.username == "sigep"
    assert client.password == "XXXXXX"


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@mock.patch('zeep.client.OperationProxy.__call__')
def test_client_timeout_error(mock_soap_client, client):
    mock_soap_client.side_effect = ConnectTimeout()
    with pytest.raises(ConnectTimeoutError):
        client.find_zipcode(ZipCode("70002-900"))


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_client_authentication_error(client):
    with pytest.raises(AuthenticationError):
        client.get_user(contract_number="9911222777", posting_card_number="0056789123")


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_client_canceled_posting_card_error(client):
    with pytest.raises(CanceledPostingCardError):
        client.get_user(contract_number="9911222777", posting_card_number="0057018901")


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_client_nonexistent_posting_card_error(client):
    with pytest.raises(NonexistentPostingCardError):
        client.get_user(contract_number="9911222777", posting_card_number="4444444444")


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_user(client):
    user = client.get_user(contract_number="9911222777", posting_card_number="0056789123")

    assert user.name == "ECT"
    assert user.federal_tax_number == "34028316000103"
    assert user.state_tax_number == "0733382100116"
    assert user.status_number == 1
    assert len(user.contracts) == 1

    contract = user.contracts[0]
    assert len(contract.posting_cards) == 1


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_find_zip_code(client):
    zip_address = client.find_zipcode(ZipCode("70002-900"))

    assert zip_address.id == 0
    assert zip_address.zip_code == "70002900"
    assert zip_address.state == "DF"
    assert zip_address.city == "Brasília"
    assert zip_address.district == "Asa Norte"
    assert zip_address.address == "SBN Quadra 1 Bloco A"
    assert zip_address.complements == []


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_verify_service_availability(client, posting_card):
    status = client.verify_service_availability(posting_card, SERVICE_SEDEX10, "82940150", "01310000")
    assert status is True


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_posting_card_status(client, posting_card):
    status = client.get_posting_card_status(posting_card)
    assert status == PostingCard.ACTIVE


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_request_tracking_codes(client, user):
    result = client.request_tracking_codes(user, Service.get(SERVICE_SEDEX), quantity=10)
    assert len(result) == 10
    assert len(result[0].code) == 13


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_generate_verification_digit(client):
    result = client.generate_verification_digit(["DL74668653 BR"])
    assert result[0] == 6


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_close_posting_list(client, posting_card, posting_list, shipping_label):
    shipping_label.posting_card = posting_card
    posting_list.add_shipping_label(shipping_label)
    posting_list = client.close_posting_list(posting_list, posting_card)
    assert posting_list.number is not None
    assert posting_list.closed


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@mock.patch('zeep.client.OperationProxy.__call__')
def test_client_close_posting_list_error(
    mock_soap_client,
    client,
    posting_card,
    posting_list,
    shipping_label,
):
    mock_soap_client.side_effect = Fault(
        'A PLP não será fechada , o(s) objeto(s) [PP40233163BR] já estão '
        'vinculados em outra PLP!'
    )
    shipping_label.posting_card = posting_card
    posting_list.add_shipping_label(shipping_label)
    with pytest.raises(ClosePostingListError):
        client.close_posting_list(posting_list, posting_card)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_codes_events(client):
    result = client.get_tracking_code_events(["FJ064849483BR", "DU477828695BR"])

    assert len(result) == 2
    assert result[0] != result[1]

    assert isinstance(result[0], TrackingCode)
    assert result[0].code in ("FJ064849483BR", "DU477828695BR")

    assert isinstance(result[1], TrackingCode)
    assert result[1].code in ("FJ064849483BR", "DU477828695BR")


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_code_events(client):
    result = client.get_tracking_code_events("FJ064849483BR")

    assert isinstance(result[0], TrackingCode)
    assert result[0].code == "FJ064849483BR"


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_code_events_without_city_field(client):
    result = client.get_tracking_code_events("PJ651329640BR")

    assert isinstance(result[0], TrackingCode)
    assert result[0].code == "PJ651329640BR"
    assert result[0].events[0].city == ""


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_code_with_no_verification_digitevents(client):
    result = client.get_tracking_code_events("FJ06484948BR")

    assert isinstance(result[0], TrackingCode)
    assert result[0].code == "FJ064849483BR"


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_code_object_not_found_by_correios(client):
    tracking_code = client.get_tracking_code_events("DU05508759BR")[0]
    assert tracking_code.events

    event = tracking_code.events[0]
    assert isinstance(event, NotFoundTrackingEvent)
    assert event.timestamp
    assert event.status.type == "ERROR"
    assert event.status.status == 0


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_get_tracking_codes_events_over_limit(client):
    codes = ["DU05508759BR"] * 51
    with pytest.raises(TrackingCodesLimitExceededError):
        client.get_tracking_code_events(codes)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_builder_posting_card_status():
    builder = ModelBuilder()
    assert builder.build_posting_card_status("Normal") == PostingCard.ACTIVE
    assert builder.build_posting_card_status("Cancelado") == PostingCard.CANCELLED


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_posting_list_serialization(posting_list, shipping_label):
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)
    assert xml.startswith(b'<?xml version="1.0" encoding="ISO-8859-1"?><correioslog>')
    assert b"<codigo_servico_adicional>019</codigo_servico_adicional>" not in xml
    assert b"<valor_declarado>10,29</valor_declarado>" not in xml


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_posting_list_serialization_with_crazy_utf8_character(posting_list, shipping_label):
    shipping_label.receiver.neighborhood = 'Olho D’Água'
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)
    assert xml.startswith(b'<?xml version="1.0" encoding="ISO-8859-1"?><correioslog>')


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_declared_value(posting_list, shipping_label):
    shipping_label.extra_services.append(ExtraService.get(EXTRA_SERVICE_VD))
    shipping_label.real_value = 10.29
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)
    assert shipping_label.service == Service.get(SERVICE_PAC)
    assert b"<codigo_servico_adicional>019</codigo_servico_adicional>" in xml
    assert b"<valor_declarado>18,50</valor_declarado>" in xml


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_fail_empty_posting_list_serialization(posting_list):
    serializer = PostingListSerializer()
    with pytest.raises(PostingListSerializerError):
        serializer.get_document(posting_list)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_fail_closed_posting_list_serialization(posting_list, shipping_label):
    posting_list.add_shipping_label(shipping_label)
    posting_list.close_with_id(number=12345)

    serializer = PostingListSerializer()
    with pytest.raises(PostingListSerializerError):
        serializer.get_document(posting_list)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_limit_size_city_name(posting_list, shipping_label):
    shipping_label.receiver.city = 'Porto Alegre (Rio Grande do Sul)'
    shipping_label.sender.city = 'Santa Maria (Rio Grande do Sul)'
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)

    assert b"<cidade_destinatario><![CDATA[Porto Alegre (Rio Grande do Su]]></cidade_destinatario>" in xml
    assert b"<cidade_remetente><![CDATA[Santa Maria (Rio Grande do Sul]]></cidade_remetente>" in xml


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freights(client, posting_card, package):
    freights = client.calculate_freights(posting_card, [SERVICE_SEDEX, SERVICE_PAC], "07192100", "80030001", package)
    assert len(freights) == 2

    freight = freights[0]
    assert freight.error_code == 0
    assert not freight.error_message
    assert freight.service == SERVICE_SEDEX
    assert freight.delivery_time.days == 1
    assert freight.total == Decimal("23.75")
    assert freight.saturday is True
    assert freight.home is True

    freight = freights[1]
    assert freight.error_code == 0
    assert not freight.error_message
    assert freight.service == SERVICE_PAC
    assert freight.delivery_time.days == 6
    assert freight.total == Decimal("14.10")
    assert freight.saturday is False
    assert freight.home is True


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freights_with_extra_services(client, posting_card, package):
    freights = client.calculate_freights(
        posting_card=posting_card,
        services=[SERVICE_SEDEX],
        from_zip="07192100",
        to_zip="80030001",
        package=package,
        value="9000.00",
        extra_services=[EXTRA_SERVICE_AR, EXTRA_SERVICE_MP]
    )
    assert len(freights) == 1

    freight = freights[0]
    assert freight.service == SERVICE_SEDEX
    assert freight.total == Decimal("96.03")
    assert freight.value == Decimal("23.75")
    assert freight.declared_value == Decimal("62.48")
    assert freight.mp_value == Decimal("5.50")
    assert freight.ar_value == Decimal("4.30")


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freight_with_error(client, posting_card, package):
    package.real_weight = 80000  # invalid weight (80kg)
    freights = client.calculate_freights(posting_card, [SERVICE_SEDEX], "99999000", "99999999", package)
    assert len(freights) == 1
    assert freights[0].error_code == -4
    assert freights[0].error_message == "Peso excedido."


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_delivery_time(client):
    expected_delivery_time = 1
    delivery_time = client.calculate_delivery_time(Service.get(SERVICE_SEDEX), '07192100', '80030001')
    assert expected_delivery_time == int(delivery_time)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_delivery_time_service_not_allowed_for_path(client):
    expected_delivery_time = 0
    delivery_time = client.calculate_delivery_time(Service.get(SERVICE_PAC), '01311300', '01311300')
    assert expected_delivery_time == int(delivery_time)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freight_with_error_code_10_restricted(client, posting_card, package):
    freights = client.calculate_freights(
        posting_card=posting_card,
        services=[SERVICE_SEDEX],
        from_zip='07192100',
        to_zip='09960610',
        package=package,
        value="9000.00",
        extra_services=[EXTRA_SERVICE_AR, EXTRA_SERVICE_MP]
    )

    assert len(freights) == 1
    freight = freights[0]  # FreightResponse

    assert isinstance(freight, FreightResponse)
    assert freight.is_restricted_address()
    assert freight.error_code == FREIGHT_ERROR_FINAL_ZIPCODE_RESTRICTED
    assert freight.value != 0
    assert freight.delivery_time.days == 2
    assert freight.saturday


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freight_with_error_code_11_restricted(
    client,
    posting_card,
    package
):
    freights = client.calculate_freights(
        posting_card=posting_card,
        services=[SERVICE_SEDEX],
        from_zip='09960610',
        to_zip='04475490',
        package=package,
        value="9000.00",
        extra_services=[EXTRA_SERVICE_AR, EXTRA_SERVICE_MP]
    )

    assert len(freights) == 1

    freight = freights[0]
    assert isinstance(freight, FreightResponse)
    assert len(freights) == 1
    assert freight.is_restricted_address()
    assert freight.error_code == FREIGHT_ERROR_INITIAL_AND_FINAL_ZIPCODE_RESTRICTED
    assert freight.value != 0
    assert freight.delivery_time.days == 8
    assert freight.saturday


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freight_with_error_code_9_restricted(
    client,
    posting_card,
    package
):
    freights = client.calculate_freights(
        posting_card=posting_card,
        services=[SERVICE_SEDEX],
        from_zip='09960610',
        to_zip='04475490',
        package=package,
        value="9000.00",
        extra_services=[EXTRA_SERVICE_AR, EXTRA_SERVICE_MP]
    )

    assert len(freights) == 1

    freight = freights[0]
    assert isinstance(freight, FreightResponse)
    assert freight.is_restricted_address()
    assert freight.error_code == FREIGHT_ERROR_INITIAL_ZIPCODE_RESTRICTED
    assert freight.value != 0
    assert freight.delivery_time.days == 8
    assert freight.saturday



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps

from lxml import etree


def add_text_argument(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        text = kwargs.pop("text", None)
        cdata = kwargs.pop("cdata", None)
        element = f(*args, **kwargs)
        if cdata:
            element.text = etree.CDATA(cdata)
        elif text:
            element.text = text
        else:
            element.text = None
        return element

    return wrapper


Element = add_text_argument(etree.Element)
SubElement = add_text_argument(etree.SubElement)
tostring = etree.tostring
parse = etree.parse
XMLSchema = etree.XMLSchema



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime, timedelta, timezone

import pytest
from PIL.Image import Image

from correios.exceptions import (
    InvalidExtraServiceError,
    InvalidFederalTaxNumberError,
    InvalidRegionalDirectionError,
    InvalidUserContractError,
    MaximumDeclaredValueError,
    MinimumDeclaredValueError,
)
from correios.models.data import EXTRA_SERVICE_AR, SERVICE_PAC, SERVICE_SEDEX
from correios.models.user import (
    Contract,
    ExtraService,
    FederalTaxNumber,
    PostingCard,
    RegionalDirection,
    Service,
    StateTaxNumber,
    User,
)


def test_basic_federal_tax_number_tax_number():
    federal_tax_number = FederalTaxNumber("73119555000120")
    assert federal_tax_number.number == "73119555000120"


def test_sanitize_federal_tax_number_tax_number():
    federal_tax_number = FederalTaxNumber("73.119.555/0001-20")
    assert federal_tax_number.number == "73119555000120"


def test_compare_federal_tax_number_with_string():
    federal_tax_number = FederalTaxNumber("73.119.555/0001-20")
    assert federal_tax_number == "73.119.555/0001-20"
    assert federal_tax_number == "73119555000120"


def test_fail_invalid_federal_tax_number():
    with pytest.raises(InvalidFederalTaxNumberError):
        FederalTaxNumber("1234567890123")

    with pytest.raises(InvalidFederalTaxNumberError):
        FederalTaxNumber("123456789012345")

    with pytest.raises(InvalidFederalTaxNumberError):
        FederalTaxNumber("73119555000199")  # invalid verification digit


def test_convert_federal_tax_number_to_str():
    assert str(FederalTaxNumber("73.119.555/0001-20")) == "73119555000120"


def test_federal_tax_number_repr():
    assert repr(FederalTaxNumber("73.119.555/0001-20")) == "<FederalTaxNumber number: 73119555000120>"


def test_federal_tax_number_display():
    assert FederalTaxNumber("73119555000120").display() == "73.119.555/0001-20"


def test_basic_state_tax_number_tax_number():
    state_tax_number = StateTaxNumber("733382100116")
    assert state_tax_number.number == "733382100116"


def test_sanitize_state_tax_number_tax_number():
    state_tax_number = StateTaxNumber("733.382.100.116")
    assert state_tax_number.number == "733382100116"


def test_convert_state_tax_number_to_str():
    assert str(StateTaxNumber("733382100116")) == "733382100116"


def test_convert_state_tax_number_display():
    assert StateTaxNumber("733382100116").display() == "733382100116"


def test_state_tax_number_repr():
    assert repr(StateTaxNumber("733.382.100.116")) == "<StateTaxNumber number: 733382100116>"


def test_basic_user(valid_federal_tax_number, valid_state_tax_number):
    user = User(name="ECT",
                federal_tax_number=valid_federal_tax_number,
                state_tax_number=valid_state_tax_number,
                status_number=1)

    assert user.name == "ECT"
    assert user.federal_tax_number == "73119555000120"
    assert user.status_number == 1
    assert user.contracts == []


def test_sanitize_user_data(valid_federal_tax_number, valid_state_tax_number):
    user = User(name="    NAME WITH TRAILLING WHITESPACES      ",
                federal_tax_number=valid_federal_tax_number,
                state_tax_number=valid_state_tax_number,
                status_number="1  ")

    assert user.name == "NAME WITH TRAILLING WHITESPACES"
    assert user.status_number == 1


def test_fail_add_contract_twice(contract, valid_federal_tax_number, valid_state_tax_number):
    user = User(name="ECT",
                federal_tax_number=valid_federal_tax_number,
                state_tax_number=valid_state_tax_number,
                status_number=1)
    user.add_contract(contract)

    with pytest.raises(InvalidUserContractError):
        user.add_contract(contract)


def test_basic_contract(user):
    contract = Contract(
        user=user,
        number=9911222777,
        regional_direction=10)

    assert contract.number == 9911222777
    assert contract.regional_direction.number == 10
    assert contract.posting_cards == []
    assert str(contract) == "9911222777"
    assert repr(contract) == "<Contract number=9911222777>"


def test_sanitize_contract_data(user):
    contract = Contract(
        user=user,
        number="9911222777  ",
        regional_direction="   10",
    )
    contract.customer_code = "279311"
    contract.status_code = "A"
    contract.start_date = "2014-05-09 00:00:00-03:00"
    contract.end_date = "2018-05-16 00:00:00-03:00"

    assert contract.number == 9911222777
    assert contract.regional_direction.number == 10
    assert contract.customer_code == 279311
    assert contract.status_code == "A"
    assert contract.start_date == datetime(2014, 5, 9, 0, 0, tzinfo=timezone(timedelta(hours=-3)))
    assert contract.end_date == datetime(2018, 5, 16, 0, 0, tzinfo=timezone(timedelta(hours=-3)))


def test_basic_posting_card(contract):
    posting_card = PostingCard(
        contract=contract,
        number=56789123,
        administrative_code=8888888,
    )

    assert posting_card.number == "0056789123"
    assert posting_card.administrative_code == "08888888"
    assert posting_card.get_contract_number() == 9911222777
    assert str(posting_card) == "0056789123"
    assert repr(posting_card) == "<PostingCard number='0056789123', contract=9911222777>"


def test_sanitize_posting_card_data(contract):
    posting_card = PostingCard(
        contract=contract,
        number="0056789123",
        administrative_code=8888888,
    )
    posting_card.start_date = "2014-05-09 00:00:00-03:00"
    posting_card.end_date = "2018-05-16 00:00:00-03:00"
    posting_card.status = "01"
    posting_card.status_code = "I"
    posting_card.unit = "08        "

    assert posting_card.number == "0056789123"
    assert posting_card.administrative_code == "08888888"
    assert posting_card.start_date == datetime(2014, 5, 9, 0, 0, tzinfo=timezone(timedelta(hours=-3)))
    assert posting_card.end_date == datetime(2018, 5, 16, 0, 0, tzinfo=timezone(timedelta(hours=-3)))
    assert posting_card.status == 1
    assert posting_card.status_code == "I"
    assert posting_card.unit == 8


def test_basic_service():
    service = Service(
        code=40215,
        id=104707,
        description="SEDEX 10",
        category="SERVICO_COM_RESTRICAO",
        symbol="premium",
        max_weight=10000,
    )

    assert service.id == 104707
    assert service.code == '40215'
    assert service.display_name == "SEDEX 10"
    assert service.description == "SEDEX 10"
    assert service.category == "SERVICO_COM_RESTRICAO"
    assert service.get_symbol_filename().endswith("/premium.gif")
    assert service.get_symbol_filename("png").endswith("/premium.png")
    assert isinstance(service.symbol_image, Image)
    assert service.max_weight == 10000


def test_sanitize_service():
    service = Service(
        code="40215                    ",
        id=104707,
        description="SEDEX 10                      ",
        category="SERVICO_COM_RESTRICAO",
    )

    assert service.id == 104707
    assert service.code == '40215'
    assert service.description == "SEDEX 10"
    assert service.category == "SERVICO_COM_RESTRICAO"


def test_service_getter():
    service = Service.get(40215)
    assert service.id == 104707
    assert service.code == '40215'
    assert service.description == "SEDEX 10"
    assert service.category == "SERVICO_COM_RESTRICAO"
    assert Service.get(service) == service


def test_fail_invalid_declared_value_in_sedex():
    service = Service.get(SERVICE_SEDEX)
    with pytest.raises(MaximumDeclaredValueError):
        service.validate_declared_value(service.max_declared_value + 1)

    with pytest.raises(MinimumDeclaredValueError):
        service.validate_declared_value(service.min_declared_value - 1)


def test_fail_invalid_declared_value_in_pac():
    service = Service.get(SERVICE_PAC)
    with pytest.raises(MaximumDeclaredValueError):
        service.validate_declared_value(service.max_declared_value + 1)

    with pytest.raises(MinimumDeclaredValueError):
        service.validate_declared_value(service.min_declared_value - 1)


def test_basic_extra_service():
    extra_service = ExtraService(1, "AR", "Aviso de Recebimento")
    assert extra_service.number == 1
    assert extra_service.code == "AR"
    assert extra_service.name == "Aviso de Recebimento"
    assert extra_service.display_on_label is True
    assert repr(extra_service) == "<ExtraService number=1, code='AR'>"
    assert extra_service == 1


def test_extra_service_sanitize_code():
    extra_service = ExtraService(1, "ar", "Aviso de Recebimento")
    assert extra_service.code == "AR"


@pytest.mark.parametrize("number,code,name", (
    (0, "XY", "Invalid Number"),
    (1, "XYZ", "Invalid Code"),
    (1, "", "Invalid Code"),
    (1, "XY", ""),  # Invalid Name
))
def test_fail_extra_service_invalid_data(number, code, name):
    with pytest.raises(InvalidExtraServiceError):
        ExtraService(number, code, name)


def test_fail_get_unknown_service():
    with pytest.raises(KeyError):
        ExtraService.get(0)


@pytest.mark.parametrize("number,extra_service_code", (
    (1, "AR"),
    (2, "MP"),
    (19, "VD"),
    (25, "RR"),
    (ExtraService.get(EXTRA_SERVICE_AR), "AR"),
))
def test_extra_service_getter(number, extra_service_code):
    assert ExtraService.get(number).code == extra_service_code


def test_basic_regional_direction():
    regional_direction = RegionalDirection(1, "AC", "AC - ADMINISTRAÇAO CENTRAL")
    assert regional_direction.number == 1
    assert regional_direction.code == "AC"
    assert regional_direction.name == "AC - ADMINISTRAÇAO CENTRAL"
    assert repr(regional_direction) == "<RegionalDirection number=1, code='AC'>"


def test_regional_direction_sanitize_code():
    regional_direction = RegionalDirection(1, "ac", "AC - ADMINISTRAÇAO CENTRAL")
    assert regional_direction.code == "AC"


@pytest.mark.parametrize("number,code,name", (
    (0, "XY", "Invalid Number"),
    (1, "", "Invalid Code"),
    (1, "XY", ""),  # Invalid Name
))
def test_fail_regional_direction_invalid_data(number, code, name):
    with pytest.raises(InvalidRegionalDirectionError):
        RegionalDirection(number, code, name)


def test_fail_get_unknown_regional_direction():
    with pytest.raises(KeyError):
        RegionalDirection.get(0)


def test_regional_direction_getter():
    regional_direction = RegionalDirection.get(1)
    assert regional_direction.number == 1
    assert regional_direction.code == "AC"
    assert regional_direction.name == "AC - ADMINISTRAÇAO CENTRAL"
    assert RegionalDirection.get(regional_direction) == regional_direction



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
from datetime import datetime, timedelta
from decimal import Decimal
from typing import Dict, List, Optional, Tuple, Union  # noqa: F401

from PIL import Image

from .. import exceptions
from ..utils import get_resource_path, to_decimal
from .address import Address, ZipCode
from .data import (
    FREIGHT_ERROR_FINAL_ZIPCODE_RESTRICTED,
    FREIGHT_ERROR_INITIAL_AND_FINAL_ZIPCODE_RESTRICTED,
    FREIGHT_ERROR_INITIAL_ZIPCODE_RESTRICTED,
    INSURANCE_PERCENTUAL_COST,
    INSURANCE_VALUE_THRESHOLD_PAC,
    INSURANCE_VALUE_THRESHOLD_SEDEX,
    SERVICE_PAC,
    SERVICE_SEDEX,
    TRACKING_EVENT_TYPES,
    TRACKING_STATUS,
)
from .user import Contract  # noqa: F401
from .user import ExtraService, PostingCard, Service

TRACKING_CODE_SIZE = 13
TRACKING_CODE_NUMBER_SIZE = 8
TRACKING_CODE_PREFIX_SIZE = 2
TRACKING_CODE_SUFFIX_SIZE = 2

IATA_COEFICIENT = 6.0

VOLUMETRIC_WEIGHT_THRESHOLD = 10000  # g

MIN_WIDTH, MAX_WIDTH = 11, 105  # cm
MIN_HEIGHT, MAX_HEIGHT = 2, 105  # cm
MIN_LENGTH, MAX_LENGTH = 16, 105  # cm
MIN_DIAMETER, MAX_DIAMETER = 5, 91  # cm
MIN_CYLINDER_LENGTH, MAX_CYLINDER_LENGTH = 18, 105  # cm
MIN_SIZE, MAX_SIZE = 29, 200  # cm
MIN_CYLINDER_SIZE, MAX_CYLINDER_SIZE = 28, 200  # cm

MAX_MECHANIZABLE_PACKAGE_SIZE = 70  # cm
NON_MECHANIZABLE_COST = Decimal('20.0')


INSURANCE_VALUE_THRESHOLDS = {
    Service.get(SERVICE_PAC).code: INSURANCE_VALUE_THRESHOLD_PAC,
    Service.get(SERVICE_SEDEX).code: INSURANCE_VALUE_THRESHOLD_SEDEX
}


class EventStatus:
    def __init__(self,
                 event_type: str,
                 status_code: Union[str, int]) -> None:
        event_type = event_type.upper()
        status_code = int(status_code)
        event_status_data = self._get_event_status_data(event_type, status_code)

        self.type = event_type
        self.status = status_code
        self.category = event_status_data[0]
        self.description = event_status_data[1]
        self.detail = event_status_data[2]
        self.action = event_status_data[3]

    def _get_event_status_data(self, event_type, status_code):
        if event_type not in TRACKING_EVENT_TYPES:
            raise exceptions.InvalidEventStatusError("{} is not valid".format(event_type))

        try:
            return TRACKING_STATUS[event_type, status_code]
        except KeyError:
            raise exceptions.InvalidEventStatusError("{} is not valid".format(event_type))

    @property
    def display_event_type(self):
        return TRACKING_EVENT_TYPES[self.type]

    def __str__(self):
        return '({}, {})'.format(self.type, self.status)

    def __repr__(self):
        return '<EventStatus({!r}, {!r})>'.format(self.type, self.status)


class ErrorEventStatus(EventStatus):
    def __init__(self):
        super().__init__("ERROR", 0)


class TrackingEvent:
    timestamp_format = "%d/%m/%Y %H:%M"

    def __init__(self,
                 timestamp: datetime,
                 status: Union[Tuple[str, Union[str, int]], EventStatus],
                 location_zip_code: Union[str, ZipCode] = "",
                 location: str = "",
                 receiver: str = "",
                 city: str = "",
                 state: str = "",
                 document: str = "",
                 comment: str = "",
                 description: str = "",
                 details: str = "",
                 ) -> None:
        self.timestamp = timestamp
        self.location = location
        self.receiver = receiver
        self.city = city
        self.state = state
        self.document = document
        self.comment = comment
        self.description = description
        self.details = details

        if location_zip_code:
            location_zip_code = ZipCode.create(location_zip_code)
        self.location_zip_code = location_zip_code

        if isinstance(status, tuple):
            status = EventStatus(*status)
        self.status = status

    def __str__(self):
        return '{} - {} - {}/{}'.format(self.description, self.location, self.city, self.state)

    def __repr__(self):
        timestamp = self.timestamp.strftime(self.timestamp_format)
        return '<TrackingEvent({!s}, {!s})>'.format(self.status, timestamp)


class NotFoundTrackingEvent(TrackingEvent):
    def __init__(self, timestamp: datetime, comment) -> None:
        super().__init__(timestamp=timestamp, status=ErrorEventStatus(), comment=comment)


class TrackingCode:
    def __init__(self, code: str) -> None:
        self.prefix = code[:2].upper()
        self.number = "".join(d for d in code[2:10] if d.isdigit())
        self.suffix = code[-2:].upper()
        self._digit = None

        if len(code) == TRACKING_CODE_SIZE and code[10:11] != " ":
            self._digit = int(code[10:11])

        self._validate()

        # filled by tracking service
        self.category = None  # type: Optional[str]
        self.name = None  # type: Optional[str]
        self.initials = None  # type: Optional[str]
        self.events = []  # type: List[TrackingEvent]

    def _validate(self):
        if len(self.prefix) != TRACKING_CODE_PREFIX_SIZE or not self.prefix.isalpha():
            raise exceptions.InvalidTrackingCodeError("Invalid tracking code prefix {}".format(self.prefix))

        if len(self.suffix) != TRACKING_CODE_SUFFIX_SIZE or not self.suffix.isalpha():
            raise exceptions.InvalidTrackingCodeError("Invalid tracking code suffix {}".format(self.suffix))

        if len(self.number) != TRACKING_CODE_NUMBER_SIZE or not self.number.isnumeric():
            raise exceptions.InvalidTrackingCodeError("Invalid tracking code number {}".format(self.number))

        if self._digit is not None and self._digit != self.calculate_digit(self.number):
            raise exceptions.InvalidTrackingCodeError("Invalid tracking code number {} or digit {} (must be {})".format(
                self.number,
                self._digit,
                self.calculate_digit(self.number))
            )

    @classmethod
    def create(cls, tracking_code: Union[str, 'TrackingCode']):
        if isinstance(tracking_code, cls):
            return tracking_code
        return cls(tracking_code)

    @classmethod
    def calculate_digit(cls, number: str) -> int:
        numbers = [int(c) for c in number if c.isdigit()]

        multipliers = [8, 6, 4, 2, 3, 5, 9, 7]
        mod = sum(multipliers[i] * digit for i, digit in enumerate(numbers)) % 11

        if not mod:
            return 5

        if mod == 1:
            return 0

        return 11 - mod

    @classmethod
    def create_range(cls, start: Union[str, 'TrackingCode'], end: Union[str, 'TrackingCode']):
        if not isinstance(start, TrackingCode):
            start = TrackingCode(start)

        if not isinstance(end, TrackingCode):
            end = TrackingCode(end)

        if start.prefix != end.prefix:
            raise exceptions.InvalidTrackingCodeError(
                "Different tracking code prefixes: {} != {}".format(start.prefix, end.prefix)
            )

        if start.suffix != end.suffix:
            raise exceptions.InvalidTrackingCodeError(
                "Different tracking code suffixes: {} != {}".format(start.suffix, end.suffix)
            )

        start_number = int(start.number)
        end_number = int(end.number)

        if start_number > end_number:
            raise exceptions.InvalidTrackingCodeError(
                "Invalid range numbers: {} > {}".format(start_number, end_number)
            )

        code_range = range(int(start.number), int(end.number) + 1)
        return [TrackingCode(start.prefix + "{:08}".format(n) + start.suffix) for n in code_range]

    @property
    def digit(self):
        if self._digit is None:
            self._digit = self.calculate_digit(self.number)
        return self._digit

    @property
    def code(self):
        return self.prefix + self.number + str(self.digit) + self.suffix

    @property
    def nodigit(self):
        return "{}{} {}".format(self.prefix, self.number, self.suffix)

    @property
    def short(self):
        return "{}{}{}".format(self.prefix, self.number, self.suffix)

    @property
    def splitted(self):
        code = self.code
        return "{!s} {!s} {!s} {!s} {!s}".format(code[:2], code[2:5], code[5:8], code[8:11], code[11:])

    def add_event(self, event: TrackingEvent):
        self.events.append(event)

    def __str__(self):
        return self.code

    def __repr__(self):
        return "<TrackingCode code={!r}>".format(self.code)


class Package:
    TYPE_ENVELOPE = 1  # type: int
    TYPE_BOX = 2  # type: int
    TYPE_CYLINDER = 3  # type: int

    freight_package_types = {
        TYPE_BOX: 1,
        TYPE_CYLINDER: 2,
        TYPE_ENVELOPE: 3,
    }  # type: Dict[int, int]

    def __init__(self,
                 package_type: int = TYPE_BOX,
                 width: Union[float, int] = 0,  # cm
                 height: Union[float, int] = 0,  # cm
                 length: Union[float, int] = 0,  # cm
                 diameter: Union[float, int] = 0,  # cm
                 weight: Union[float, int] = 0,  # g
                 sequence=(1, 1),
                 service: Optional[Union[Service, str, int]] = None) -> None:

        if service:
            service = Service.get(service)

        Package.validate(package_type, width, height, length, diameter, service, weight)

        if len(sequence) != 2 or sequence[0] > sequence[1]:
            raise exceptions.InvalidPackageSequenceError("Package must be a tuple with 2 elements: (number, total)")

        self.package_type = package_type

        self.real_width = width  # cm
        self.real_height = height  # cm
        self.real_length = length  # cm
        self.real_diameter = diameter  # cm
        self.real_weight = weight  # g
        self.sequence = sequence
        self.service = service

    @property
    def width(self) -> int:
        return max(MIN_WIDTH, int(math.ceil(self.real_width)))

    @width.setter
    def width(self, width):
        Package._validate_dimension("width", width, MAX_WIDTH)
        self.real_width = width

    @property
    def height(self) -> int:
        return max(MIN_HEIGHT, int(math.ceil(self.real_height)))

    @height.setter
    def height(self, height):
        Package._validate_dimension("height", height, MAX_HEIGHT)
        self.real_height = height

    @property
    def length(self) -> int:
        return max(MIN_LENGTH, int(math.ceil(self.real_length)))

    @length.setter
    def length(self, length):
        Package._validate_dimension("length", length, MAX_LENGTH)
        self.real_length = length

    @property
    def diameter(self) -> int:
        if self.package_type != Package.TYPE_CYLINDER:
            return 0
        return max(MIN_DIAMETER, int(math.ceil(self.real_diameter)))

    @diameter.setter
    def diameter(self, diameter):
        Package._validate_dimension("diameter", diameter, MAX_DIAMETER)
        self.real_diameter = diameter

    @property
    def weight(self) -> int:
        return int(math.ceil(self.real_weight))

    @weight.setter
    def weight(self, weight):
        Package._validate_weight(weight, self.service)
        self.real_weight = weight

    @property
    def volumetric_weight(self) -> int:
        return Package.calculate_volumetric_weight(self.width, self.height, self.length)

    @property
    def posting_list_volumetric_weight(self) -> Decimal:
        return Decimal("0.00")

    @property
    def posting_weight(self) -> int:
        return Package.calculate_posting_weight(self.weight, self.volumetric_weight)

    @property
    def freight_package_type(self) -> int:
        """
        SIGEP API and FreightResponse API different codes to identify package types:

        SIGEP | Freight | Type
        ------+---------+----------
          1   |    3    | Envelope
          2   |    1    | Box
          3   |    2    | Cylinder
        """
        return self.freight_package_types[self.package_type]

    @property
    def is_mechanizable(self) -> bool:
        if self.package_type == Package.TYPE_CYLINDER:
            return False
        return max(self.width, self.height, self.length) <= MAX_MECHANIZABLE_PACKAGE_SIZE

    @property
    def non_mechanizable_cost(self):
        return Decimal('0.0') if self.is_mechanizable else NON_MECHANIZABLE_COST

    @classmethod
    def calculate_volumetric_weight(cls, width, height, length) -> int:
        return int(math.ceil((width * height * length) / IATA_COEFICIENT))

    @classmethod
    def calculate_posting_weight(cls, weight, volumetric_weight) -> int:
        if volumetric_weight <= VOLUMETRIC_WEIGHT_THRESHOLD:
            return weight
        return int(math.ceil(max(volumetric_weight, weight)))

    @classmethod
    def calculate_insurance(cls,
                            per_unit_value: Union[int, float, Decimal],
                            service: Union[Service, int, str],
                            quantity: int = 1) -> Decimal:
        value = Decimal("0.00")
        per_unit_value = Decimal(per_unit_value)
        service_code = Service.get(service).code
        insurance_value_threshold = INSURANCE_VALUE_THRESHOLDS.get(service_code, per_unit_value)

        if per_unit_value > insurance_value_threshold:
            value = (per_unit_value - insurance_value_threshold) * INSURANCE_PERCENTUAL_COST

        return to_decimal(value * quantity)

    @classmethod
    def validate(cls,
                 package_type: int,
                 width: Union[float, int] = 0,
                 height: Union[float, int] = 0,
                 length: Union[float, int] = 0,
                 diameter: Union[float, int] = 0,
                 service: Optional[Union[Service, str, int]] = None,
                 weight: Union[float, int] = 0) -> None:

        width = int(math.ceil(width))
        height = int(math.ceil(height))
        length = int(math.ceil(length))
        diameter = int(math.ceil(diameter))
        weight = int(math.ceil(weight))

        if service:
            service = Service.get(service)

        Package._validate_weight(weight, service)

        if package_type == Package.TYPE_ENVELOPE:
            if any([width, height, length, diameter]):
                raise exceptions.InvalidPackageDimensionsError(
                    "Invalid dimensions: {}x{}x{}".format(width, height, length)
                )
            return

        if package_type == Package.TYPE_BOX:
            if diameter:
                raise exceptions.InvalidPackageDimensionsError(
                    "Package does not use diameter: {}".format(diameter)
                )

            cls._validate_dimension("width", width, MAX_WIDTH)
            cls._validate_dimension("height", height, MAX_HEIGHT)
            cls._validate_dimension("length", length, MAX_LENGTH)
            cls._validate_dimension("sum of dimensions", width + height + length, MAX_SIZE)
            return

        # Volume.TYPE_CYLINDER
        if width or height:
            raise exceptions.InvalidPackageDimensionsError(
                "Cylinder does not use width/height: {}x{}".format(width, height)
            )

        cls._validate_dimension("length", length, MAX_CYLINDER_LENGTH)
        cls._validate_dimension("diameter", diameter, MAX_DIAMETER)
        cls._validate_dimension("cylinder size", length + 2 * diameter, MAX_CYLINDER_SIZE)

    @classmethod
    def _validate_dimension(cls, name, value, maximum):
        msg = "Invalid {} (range 0~{}): {}".format(name, maximum, value)
        if value <= 0:
            raise exceptions.InvalidMinPackageDimensionsError(msg)

        if value > maximum:
            raise exceptions.InvalidMaxPackageDimensionsError(msg)

    @classmethod
    def _validate_weight(cls, weight, service: Optional[Union[Service, str, int]] = None) -> None:
        if weight <= 0:
            raise exceptions.InvalidMinPackageWeightError("Invalid weight {!r}g".format(weight))

        if not service:
            return

        service = Service.get(service)

        if service.max_weight is None:
            return

        if weight > service.max_weight:
            message = "Max weight exceeded for service {!r}: {!r}g (max. {!r}g)".format(
                weight,
                str(service),
                service.max_weight,
            )
            raise exceptions.InvalidMaxPackageWeightError(message)


class ShippingLabel:
    variable_data_identifier = 51  # Variable data identifier for package
    invoice_template = "{!s}"
    contract_number_template = "{!s}"
    order_template = "{!s}"
    service_name_template = "{!s}"
    package_template = "{!s}/{!s}"
    weight_template = "{!s}g"
    receipt_template = ("Recebedor: ___________________________________________<br/>"
                        "Assinatura: __________________ Documento: _______________")
    sender_header = "DESTINATÁRIO"
    carrier_logo = str(get_resource_path("carrier_logo_bw.png"))
    receiver_data_template = ("{receiver.label_name!s:>.50}<br/>"
                              "{receiver.label_address!s:>.95}<br/>"
                              "<b>{receiver.zip_code_display}</b> {receiver.city}/{receiver.state}")

    sender_data_template = ("<b>Remetente:</b> {sender.label_name!s:>.40}<br/>"
                            "{sender.label_address!s:>.95}<br/>"
                            "<b>{sender.zip_code_display}</b> {sender.city}-{sender.state}")

    def __init__(self,
                 posting_card: PostingCard,
                 sender: Address,
                 receiver: Address,
                 service: Union[Service, int],
                 tracking_code: Union[TrackingCode, str],
                 package: Package,
                 extra_services: Optional[List[Union[ExtraService, int]]] = None,
                 logo: Optional[Union[str, Image.Image]] = None,
                 order: Optional[str] = "",
                 invoice_number: str = "",
                 invoice_series: str = "",
                 invoice_type: str = "",
                 value: Decimal = Decimal("0.00"),
                 billing: Decimal = Decimal("0.00"),
                 text: str = "",
                 latitude: float = 0.0,
                 longitude: float = 0.0) -> None:

        if sender == receiver:
            raise exceptions.InvalidAddressesError("Sender and receiver cannot be the same")

        if logo is None:
            logo = str(get_resource_path("default_logo.png"))

        if isinstance(logo, str):
            logo = Image.open(logo)

        self.posting_card = posting_card
        self.sender = sender
        self.receiver = receiver
        self.service = Service.get(service)
        self.tracking_code = TrackingCode.create(tracking_code)
        self.package = package
        self.logo = logo
        self.order = order
        self.invoice_number = invoice_number
        self.invoice_series = invoice_series
        self.invoice_type = invoice_type
        self.real_value = value
        self.billing = billing
        self.text = text
        self.latitude = latitude
        self.longitude = longitude
        self.carrier_logo = Image.open(self.carrier_logo)

        self.extra_services = self.service.default_extra_services[:]
        if extra_services:
            self.add_extra_services(extra_services)

        self.posting_list = None  # type: Optional[PostingList]
        self.posting_list_group = 0

    def __repr__(self):
        return "<ShippingLabel tracking={!r}>".format(str(self.tracking_code))

    def add_extra_services(self, extra_services: List[Union["ExtraService", int]]):
        for extra_service in extra_services:
            self.add_extra_service(extra_service)

    def add_extra_service(self, extra_service: Union["ExtraService", int]):
        extra_service = ExtraService.get(extra_service)
        if extra_service.is_declared_value():
            self.service.validate_declared_value(self.value)
        self.extra_services.append(extra_service)

    @property
    def value(self) -> Decimal:
        return max(self.service.min_declared_value, self.real_value)

    @property
    def symbol(self):
        return self.service.symbol_image

    @property
    def contract(self):
        return self.posting_card.contract

    @property
    def posting_weight(self):
        return self.package.posting_weight

    def get_order(self):
        return self.order_template.format(self.order)

    def get_invoice(self):
        return self.invoice_template.format(self.invoice_number)

    def get_contract_number(self):
        return self.contract_number_template.format(self.posting_card.get_contract_number())

    def get_service_name(self):
        return self.service_name_template.format(self.service.display_name)

    def get_package_sequence(self):
        return self.package_template.format(*self.package.sequence)

    def get_weight(self):
        return self.weight_template.format(self.package.weight)

    def get_symbol_filename(self, extension="gif"):
        return self.service.get_symbol_filename(extension)

    def get_tracking_code(self):
        return self.tracking_code.splitted

    def get_receiver_data(self):
        return self.receiver_data_template.format(receiver=self.receiver)

    def get_sender_data(self):
        return self.sender_data_template.format(sender=self.sender)

    def _get_extra_service_info(self) -> str:
        extra_services_numbers = ["00" for _ in range(6)]
        for i, extra_service in enumerate(self.extra_services[:6]):
            extra_services_numbers[i] = "{:02d}".format(extra_service.number)
        return "".join(extra_services_numbers)

    def get_datamatrix_info(self):
        receiver_number = self.receiver.number
        if receiver_number.isnumeric():
            receiver_number = receiver_number.rjust(5, "0")
        else:
            receiver_number = receiver_number.rjust(5)

        parts = [
            "{!s:>08}".format(self.receiver.zip_code),
            "{}".format(self.receiver.zip_complement.rjust(5, "0")),
            "{!s:>08}".format(self.sender.zip_code),
            "{}".format(self.sender.zip_complement.rjust(5, "0")),
            "{!s:>01}".format(self.receiver.zip_code.digit),
            "{!s:>02}".format(self.variable_data_identifier),
            "{!s:>13}".format(self.tracking_code),
            "{!s:>12}".format(self._get_extra_service_info()),
            "{!s:>010}".format(self.posting_card.number),
            "{!s:>05}".format(self.service.code),
            "{!s:>02}".format(self.posting_list_group),
            "{}".format(receiver_number),
            "{!s:<20}".format(self.receiver.complement[:20]),
            "{!s:>05}".format(0 if self.value is None else int(self.value * 100)),
            "{}".format(str(self.receiver.phone)[:12].rjust(12, "0") or "0" * 12),
            "{:+010.6f}".format(self.latitude),
            "{:+010.6f}".format(self.longitude),
            "|",
            "{!s:<30}".format(self.text[:30]),
        ]
        return "".join(parts)

    def __contains__(self, extra_service: ExtraService):
        return extra_service in self.extra_services


class PostingList:
    def __init__(self, custom_id: int, logo: Optional[Union[str, Image.Image]] = None) -> None:
        # will be filled by close_posting_list
        self.number = None  # type: Optional[int]

        if logo is None:
            logo = str(get_resource_path("carrier_logo.png"))

        if isinstance(logo, str):
            logo = Image.open(logo)

        self.logo = logo
        self.custom_id = custom_id
        self.shipping_labels = {}  # type: Dict[str, ShippingLabel]

        # filled by the first shipping label
        self.initial_shipping_label = None  # type: Optional[ShippingLabel]
        self.posting_card = None  # type: Optional[PostingCard]
        self.contract = None  # type: Optional[Contract]
        self.sender = None  # type: Optional[Address]

    def add_shipping_label(self, shipping_label: ShippingLabel):
        if not self.initial_shipping_label:
            self.initial_shipping_label = shipping_label
            self.posting_card = shipping_label.posting_card
            self.contract = shipping_label.contract
            self.sender = shipping_label.sender

        if shipping_label.tracking_code.short in self.shipping_labels:
            raise exceptions.PostingListError("Shipping label {!r} already in posting list".format(shipping_label))

        if shipping_label.posting_card != self.posting_card:
            raise exceptions.PostingListError(
                "Invalid posting card: {} != {}".format(shipping_label.posting_card, self.posting_card)
            )
        self.shipping_labels[shipping_label.tracking_code.short] = shipping_label
        shipping_label.posting_list = self

    def get_tracking_codes(self):
        return list(self.shipping_labels.keys())

    def close_with_id(self, number: int):
        self.number = number

    @property
    def closed(self):
        return self.number is not None


class FreightResponse:
    restricted_address_error_code = (
        FREIGHT_ERROR_INITIAL_ZIPCODE_RESTRICTED,
        FREIGHT_ERROR_FINAL_ZIPCODE_RESTRICTED,
        FREIGHT_ERROR_INITIAL_AND_FINAL_ZIPCODE_RESTRICTED,
    )

    def __init__(self,
                 service: Union[Service, int],
                 delivery_time: Union[int, timedelta],
                 value: Union[Decimal, float, int, str],
                 declared_value: Union[Decimal, float, int, str] = 0.00,
                 mp_value: Union[Decimal, float, int, str] = 0.00,
                 ar_value: Union[Decimal, float, int, str] = 0.00,
                 saturday: bool = False,
                 home: bool = False,
                 error_code: int = 0,
                 error_message: str = "") -> None:

        self.service = Service.get(service)

        if not isinstance(delivery_time, timedelta):
            delivery_time = timedelta(days=delivery_time)
        self.delivery_time = delivery_time

        if not isinstance(value, Decimal):
            value = to_decimal(value)
        self.value = value

        if not isinstance(declared_value, Decimal):
            declared_value = to_decimal(declared_value)
        self.declared_value = declared_value

        if not isinstance(mp_value, Decimal):
            mp_value = to_decimal(mp_value)
        self.mp_value = mp_value

        if not isinstance(ar_value, Decimal):
            ar_value = to_decimal(ar_value)
        self.ar_value = ar_value

        if not isinstance(saturday, bool):
            saturday = (saturday == 'S')
        self.saturday = saturday

        if not isinstance(home, bool):
            home = (home == 'S')
        self.home = home

        self.error_code = error_code
        self.error_message = error_message

    @property
    def total(self) -> Decimal:
        return self.value + self.declared_value + self.ar_value + self.mp_value

    def is_error(self):
        return self.error_code != 0

    def is_restricted_address(self):
        return self.error_code in self.restricted_address_error_code
